#ifndef _JDISP_DOMUTILS_JSH
#define _JDISP_DOMUTILS_JSH

#define DOMUTILS_BODY_PROPERTY(a) (document.body.a || document.documentElement.a)
#define _STYLE_LENGTH(name) (this.parsePx(name) || 0)

OOP_CLASS(DomUtils) = function() {};
OOP_CLASS(DomUtils).useDOM2 = false;
if (IS_OBJECT(document.implementation)) {
  if (document.implementation.hasFeature('Events', '2.0')) {
    OOP_CLASS(DomUtils).useDOM2 = true;
  }
}

OOP_METHOD(DomUtils, getNode) = function (node)
{
  if (IS_OBJECT(node)) return node;
  return document.getElementById(node);
};

OOP_METHOD(DomUtils, setHtml) = function (node, html)
{
  var n = this.getNode(node);
  if (IS_NON_NULL_OBJECT(n)) {
    n.innerHTML = html;
  }
};

OOP_METHOD(DomUtils, getEventObject) = function (ev)
{
  return (IS_NON_NULL_OBJECT(ev) ? ev : window.event);
};

OOP_METHOD(DomUtils, fixStaticPosition) = function (node)
{
  var n = this.getNode(node);
  if (!IS_NON_NULL_OBJECT(n)) return;
  var s = n.style;
  if (!IS_NON_NULL_OBJECT(s)) return;
  if (s.position === '') {
    s.position = 'relative';
    s.left = '0';
    s.top = '0';
  }
};


OOP_METHOD(DomUtils, parseEventObject) = function (ev)
{
  var ee = this.getEventObject(ev);
  var src = this.getEventSource(ee);
  var pos = this.getEventPagePosition(ee);
  var off = this.getEventOffset(ee);
  var ret = {event : ee, src : src, position: pos, offset: off};
  return ret;
};

OOP_METHOD(DomUtils, getEventSource) = function (ev)
{
  var ee = this.getEventObject(ev);
  return ((document.all) ? (ee.srcElement) : ee.target);
};

OOP_METHOD(DomUtils, getRelativePosition) = function (fromNode, toNode)
{
  var fn = this.getNode(fromNode);
  var tn = this.getNode(toNode);
  var pos = {x:0, y:0};

  var n = tn;
  var s, p;

  while (!IS_NULL(n)) {
    if (n === fn) return pos;
    p = this.getElementPosition(n);
    if (IS_NULL(p)) break;
    pos.x += p.x;
    pos.y += p.y;
    s = n.style;
    if (IS_NULL(s)) break;
    pos.x += _STYLE_LENGTH(s.marginLeft) +
             _STYLE_LENGTH(s.borderLeftWidth);
    pos.y += _STYLE_LENGTH(s.marginTop) +
             _STYLE_LENGTH(s.borderTopWidth);
    n = n.parentNode;
  }

  return null;
};

OOP_METHOD(DomUtils, getEventOffset) = function (ev, fromNode)
{
  var fn = this.getNode(fromNode);
  var ee = this.getEventObject(ev);
  var src = this.getEventSource(ee);
  var s = src.style;
  var ret = null;
  if (document.all) {
    ret = { x : (ee.offsetX + _STYLE_LENGTH(s.borderLeftWidth)),
            y : (ee.offsetY + _STYLE_LENGTH(s.borderTopWidth)) };
    if (BROWSER_IS_OPERA()) {
      ret.x += _STYLE_LENGTH(s.paddingLeft);
      ret.y += _STYLE_LENGTH(s.paddingTop);
    }
  } else {
    ret = { x : ee.layerX, y : ee.layerY };
  }
#if 0
  var a = ['offsetX', 'offsetY', 'layerX', 'layerY', 'screenX', 'screenY', 'pageX', 'pageY', 'x', 'y', 'clientX', 'clientY'];
  for (var i in a) {
    ret['_' + a[i]] = ee[a[i]];
  }
#endif

  if (IS_NON_NULL_OBJECT(fn)) {
    var rel = this.getRelativePosition(fn, src);
    if (IS_NON_NULL_OBJECT(rel)) {
      ret.x += rel.x;
      ret.y += rel.y;
    }
    ret.x -= _STYLE_LENGTH(s.borderLeftWidth);
    ret.y -= _STYLE_LENGTH(s.borderTopWidth);
  }

  return ret;
};

OOP_METHOD(DomUtils, getEventPagePosition) = function (ev)
{
  var x, y;
  var ee = this.getEventObject(ev);
  if (IS_NUMBER(ee.pageX)) {
    x = ee.pageX;
    y = ee.pageY;
  } else if (IS_NUMBER(ee.clientX)) {
    var de = document.documentElement;
    var b = document.body;
    x = ee.clientX +
        (de.scrollLeft || b.scrollLeft) -
        (de.clientLeft || 0) ;
    y = ee.clientY +
        (de.scrollTop || b.scrollTop) -
        (de.clientTop || 0) ;
  } else {
    return null;
  }

  return {x : x, y : y};
};

OOP_METHOD(DomUtils, getEventPagePosition2) = function (ev)
{
  var x = 0;
  var y = 0;
  var ee = this.getEventObject(ev);

  if (ee.pageX || ee.pageY) {
    x = ee.pageX;
    y = ee.pageY;
  } else if (ee.clientX || ee.clientY)  {
    var b = document.body;
    var de = document.documentElement;
  x = ee.clientX + b.scrollLeft + de.scrollLeft;
    y = ee.clientY + b.scrollTop + de.scrollTop;
  }
  return {x : x, y : y};
};


OOP_METHOD(DomUtils, parsePx) = function (s)
{
  if (IS_STRING(s)) {
    if (s.match(/(-?\d\d*)px/)) {
      return parseInt(RegExp.$1, 10);
    } else if (s.match(/(-?[.0-9]*)pt/)) {
      // FIXME: implement this
      return Number(RegExp.$1);
    } else {
      return undefined;
    }
  } else {
    return undefined;
  }
};

OOP_METHOD(DomUtils, getScrollPosition) = function ()
{
  var x = DOMUTILS_BODY_PROPERTY(scrollLeft);
  var y = DOMUTILS_BODY_PROPERTY(scrollTop);
  return {x:x, y:y};
};

OOP_METHOD(DomUtils, getElementPosition) = function (node)
{
  var n = this.getNode(node);
  if (IS_NON_NULL_OBJECT(n)) {
    var s = n.style;
    if (IS_NON_NULL_OBJECT(s)) {
      if (IS_NUMBER(s.posLeft)) {
        return {x: s.posLeft, y: s.posTop};
      } else {
        var x = this.parsePx(s.left);
        var y = this.parsePx(s.top);
        return {x : x, y : y};
      }
    } else {
      return null;
    }
  } else {
    return null;
  }
};

OOP_METHOD(DomUtils, setElementPosition) = function (node, x, y)
{
  var n = this.getNode(node);
  if (IS_NON_NULL_OBJECT(n)) {
    var s = n.style;
    if (IS_NON_NULL_OBJECT(s)) {
      if (IS_NUMBER(s.posLeft)) {
        s.posLeft = x;
        s.posTop  = y;
      } else {
        s.left = x + "px";
        s.top  = y + "px";
      }
    }
  }
};

OOP_METHOD(DomUtils, getPropertyValue) = function (node, name, defaultValue)
{
  var i, s;
  var n = this.getNode(node);
  if (!IS_NON_NULL_OBJECT(n)) { return undefined; }

  s = n.currentStyle;
  if (IS_NON_NULL_OBJECT(s)) {
    name = PROJECT_PREFIX.stringUtils.camelize(name);
    return this.parsePx(s[name]) || defaultValue;
  }

  s = document.defaultView.getComputedStyle(n, null);
  if (IS_NON_NULL_OBJECT(s)) {
    return this.parsePx(s.getPropertyValue(name)) || defaultValue;
  }

  return undefined;
};

OOP_METHOD(DomUtils, getElementSize) = function (node)
{
  var s, w, h, ow, oh, mt, mb, ml, mr, bt, bb, bl, br, mw, mh, bw, bh;
  var ret=null;
  var n = this.getNode(node);
  if (IS_NON_NULL_OBJECT(n)) {
    w  = this.getPropertyValue(n, 'width', 0);
    h  = this.getPropertyValue(n, 'height', 0);
    ow = n.offsetWidth || 0;
    oh = n.offsetHeight || 0;
    mt = this.getPropertyValue(n, 'margin-top', 0);
    mb = this.getPropertyValue(n, 'margin-bottom', 0);
    ml = this.getPropertyValue(n, 'margin-left', 0);
    mr = this.getPropertyValue(n, 'margin-right', 0);
    bt = this.getPropertyValue(n, 'border-top-width', 0);
    bb = this.getPropertyValue(n, 'border-bottom-width', 0);
    bl = this.getPropertyValue(n, 'border-left-width', 0);
    br = this.getPropertyValue(n, 'border-right-width', 0);
    mw = ml + mr;
    mh = mt + mb;
    bw = bl + br;
    bh = bt + bb;
  }
  ret = { width : w,
          height : h,
          offsetWidth: ow,
          offsetHeight: oh,
          outerWidth : (ow + mw),
          outerHeight : (oh + mh),
          marginTop: mt,
          marginBottom: mb,
          marginLeft: ml,
          marginRight: mr,
          marginWidth: mw,
          marginHeight: mh,
          borderWidth : bw,
          borderHeight : bh,
          borderTop: bt,
          borderBottom: bb,
          borderLeft: bl,
          borderRight: br };
  LOG_DEBUG(PROJECT_PREFIX.global.dumpVar(ret));
  return ret;
};

OOP_METHOD(DomUtils, getElementSize1) = function (node)
{
  var s, w, h, ow, oh, mt, mb, ml, mr, bt, bb, bl, br, mw, mh, bw, bh;
  var ret=null;
  var n = this.getNode(node);
  if (IS_NON_NULL_OBJECT(n)) {
    if (IS_NON_NULL_OBJECT(n.currentStyle)) {
      // IE
      s  = n.currentStyle;
      w  = this.parsePx(s.width) || 0;
      h  = this.parsePx(s.height) || 0;
      ow = n.offsetWidth || 0;
      oh = n.offsetHeight || 0;
      mt = this.parsePx(s.marginTop) || 0;
      mb = this.parsePx(s.marginBottom) || 0;
      ml = this.parsePx(s.marginLeft) || 0;
      mr = this.parsePx(s.marginRight) || 0;
      bt = this.parsePx(s.borderTopWidth) || 0;
      bb = this.parsePx(s.borderBottomWidth) || 0;
      bl = this.parsePx(s.borderLeftWidth) || 0;
      br = this.parsePx(s.borderRightWidth) || 0;
      mw = ml + mr;
      mh = mt + mb;
      bw = bl + br;
      bh = bt + bb;
    } else if (IS_NON_NULL_OBJECT(document.defaultView.getComputedStyle(n, null))) {
      s = document.defaultView.getComputedStyle(n, null);
      w  = this.parsePx(s.getPropertyValue('width')) || 0;
      h  = this.parsePx(s.getPropertyValue('height')) || 0;
      ow  = n.offsetWidth || 0;
      oh  = n.offsetHeight || 0;
      mt = this.parsePx(s.getPropertyValue('margin-top')) || 0;
      mb = this.parsePx(s.getPropertyValue('margin-bottom')) || 0;
      ml = this.parsePx(s.getPropertyValue('margin-left')) || 0;
      mr = this.parsePx(s.getPropertyValue('margin-right')) || 0;
      bt = this.parsePx(s.getPropertyValue('border-top-width')) || 0;
      bb = this.parsePx(s.getPropertyValue('border-bottom-width')) || 0;
      bl = this.parsePx(s.getPropertyValue('border-left-width')) || 0;
      br = this.parsePx(s.getPropertyValue('border-right-width')) || 0;
      mw = ml + mr;
      mh = mt + mb;
      bw = bl + br;
      bh = bt + bb;

    }
  }
      ret = { width : w,
              height : h,
              offsetWidth: ow,
              offsetHeight: oh,
              fitWidth : ow,
              fitHeight : oh,
              outerWidth : (ow + mw),
              outerHeight : (oh + mh),
              marginTop: mt,
              marginBottom: mb,
              marginLeft: ml,
              marginRight: mr,
              marginWidth: mw,
              marginHeight: mh,
              borderWidth : bw,
              borderHeight : bh,
              borderTop: bt,
              borderBottom: bb,
              borderLeft: bl,
              borderRight: br };
  PROJECT_PREFIX.logging.stdout.append(PROJECT_PREFIX.global.dumpVar(ret));
  return ret;
};

OOP_METHOD(DomUtils, getElementSizeOld) = function (node)
{
  var n = this.getNode(node);
  if (IS_NON_NULL_OBJECT(n)) {
    var s = n.style;
    if (IS_NON_NULL_OBJECT(s)) {
      var w  = this.parsePx(s.width) || 0;
      var h  = this.parsePx(s.height) || 0;
      var mt = this.parsePx(s.marginTop) || 0;
      var mb = this.parsePx(s.marginBottom) || 0;
      var ml = this.parsePx(s.marginLeft) || 0;
      var mr = this.parsePx(s.marginRight) || 0;
      var bt = this.parsePx(s.borderTopWidth) || 0;
      var bb = this.parsePx(s.borderBottomWidth) || 0;
      var bl = this.parsePx(s.borderLeftWidth) || 0;
      var br = this.parsePx(s.borderRightWidth) || 0;
      var mw = ml + mr;
      var mh = mt + mb;
      var bw = bl + br;
      var bh = bt + bb;
      PROJECT_PREFIX.logging.stdout.append(PROJECT_PREFIX.global.dumpVar({ width : w,
               height : h,
               fitWidth : (w + bw),
               fitHeight : (h + bh),
               outerWidth : (w + bw + mw),
               outerHeight : (h + bh + mh),
               marginTop: mt,
               marginBottom: mb,
               marginLeft: ml,
               marginRight: mr,
               marginWidth: mw,
               marginHeight: mh,
               borderWidth : bw,
               borderHeight : bh,
               borderTop: bt,
               borderBottom: bb,
               borderLeft: bl,
               borderRight: br }));
      return { width : w,
               height : h,
               fitWidth : (w + bw),
               fitHeight : (h + bh),
               outerWidth : (w + bw + mw),
               outerHeight : (h + bh + mh),
               marginTop: mt,
               marginBottom: mb,
               marginLeft: ml,
               marginRight: mr,
               marginWidth: mw,
               marginHeight: mh,
               borderWidth : bw,
               borderHeight : bh,
               borderTop: bt,
               borderBottom: bb,
               borderLeft: bl,
               borderRight: br };
    }
  }
  return null;
};

// OK: FF3, Opera9, Chrome1
// NG: IE6
OOP_METHOD(DomUtils, getInnerSize) = function ()
{
  var w = window.innerWidth;
  var h = window.innerHeight;
  return {width:w, height:h};
};

// OK: FF3, Opera9, Chrome1
// NG: IE6
OOP_METHOD(DomUtils, getOuterSize) = function ()
{
  var w = window.outerWidth;
  var h = window.outerHeight;
  return {width:w, height:h};
};

// IE6, FF3, Opera9, Chrome1 OK
OOP_METHOD(DomUtils, getScreenSize) = function ()
{
  var w = screen.width;
  var h = screen.height;
  return {width:w, height:h};
};

// IE6, FF3, Opera9, Chrome1 OK
OOP_METHOD(DomUtils, getAvailableSize) = function ()
{
  var w = screen.availWidth;
  var h = screen.availHeight;
  return {width:w, height:h};
};

OOP_METHOD(DomUtils, getClientSizeByBody) = function ()
{
  var w = document.body.clientWidth;
  var h = document.body.clientHeight;
  return {width:w, height:h};
};

OOP_METHOD(DomUtils, getClientSizeByDocument) = function ()
{
  var w = document.documentElement.clientWidth;
  var h = document.documentElement.clientHeight;
  return {width:w, height:h};
};

OOP_METHOD(DomUtils, getClientSize) = function ()
{
  var w, h;
  var de = document.documentElement;
  var b  = document.body;
  w = de.clientWidth || b.clientWidth;
  h = de.clientHeight || b.clientHeight;
  return {width:w, height:h};
};

OOP_METHOD(DomUtils, getScrollSizeByBody) = function ()
{
  var w = document.body.scrollWidth;
  var h = document.body.scrollHeight;
  return {width:w, height:h};
};

/*
 * To compute the client size without scroll bars:
 *  ff3 -> de.clientWidth
 *  ie6 -> de.clientWidth
 *  op9 -> b.clientWidth
 *  ch1 -> de.clientWidth
 *
 * the client size including scroll bars:
 *  ff3 -> screen.innerWidth
 *  ie6 -> n/a
 *  op9 -> screen.innerWidth
 *  ch1 -> screen.innerWidth
 */
OOP_METHOD(DomUtils, getScrollSizeByDocument) = function ()
{
  var w = document.documentElement.scrollWidth;
  var h = document.documentElement.scrollHeight;
  return {width:w, height:h};
};

OOP_METHOD(DomUtils, getScrollSize) = function ()
{
  var w = DOMUTILS_BODY_PROPERTY(scrollWidth) || window.innerWidth;
  var h = DOMUTILS_BODY_PROPERTY(scrollHeight) || window.innerHeight;
  return {width:w, height:h};
};

OOP_METHOD(DomUtils, preventDefault) = function (ev)
{
  var ee = this.getEventObject(ev);
  if (IS_FUNCTION(ee.preventDefault)) {
    ee.preventDefault();
  } else {
    ee.cancelBubble = true;
  }
};

OOP_METHOD(DomUtils, stopPropagation) = function (ev)
{
  var ee = this.getEventObject(ev);
  if (IS_FUNCTION(ee.stopPropagation)) {
    ee.stopPropagation();
  } else {
    ee.returnValue = false;
  }
};

OOP_METHOD(DomUtils, removeElementById) = function (node)
{
  var n = this.getNode(node);
  if (IS_NON_NULL_OBJECT(n))
    return n.parentNode.removeChild(n);
  return null;
};


OOP_METHOD(DomUtils, addEventListener) = function (node, type, listener, useCapture)
{
  var n = this.getNode(node);

  if (!IS_NON_NULL_OBJECT(n)) return;
  n['on' + type.toLowerCase()] = listener;
};

OOP_METHOD(DomUtils, addEventListenerDOM) = function (node, type, listener, useCapture)
{
  var n = this.getNode(node);

  if (!IS_NON_NULL_OBJECT(n)) return;

  if (OOP_CLASS(DomUtils).useDOM2) {
    n.addEventListener(type, listener, useCapture);
  } else {
    n['on' + type.toLowerCase()] = listener;
    if (n.captureEvents) {
      n.captureEvents('Event.' + type.toUpperCase());
    }
  }
};

OOP_METHOD(DomUtils, setScrollHandler) = function (f)
{
  if (document.all) {
    var de = document.documentElement;
    if (de) {
      de.onscroll = f;
    } else {
      document.body.onscroll = f;
    }
  } else {
    window.onscroll = f;
  }
};

OOP_METHOD(DomUtils, getScrollHandler) = function ()
{
  if (document.all) {
    var de = document.documentElement;
    if (de) {
      return de.onscroll;
    } else {
      return document.body.onscroll;
    }
  } else {
    return window.onscroll;
  }
};

OOP_METHOD(DomUtils, setResizeHandler) = function (f)
{
  window.onresize = f;
};

OOP_METHOD(DomUtils, getResizeHandler) = function ()
{
  return window.onresize;
};


#undef _STYLE_LENGTH

#endif /* _JDISP_DOMUTILS_JSH */
