#ifndef _JDISP_GLOBAL_JSH
#define _JDISP_GLOBAL_JSH

OOP_CLASS(Global) = function() {};

OOP_METHOD(Global, eval) = function (txt)
{
  EVAL_RAW(txt);
};

OOP_METHOD(Global, random) = function (bound)
{
  if (+bound === bound) {
    return Math.floor(Math.random() * bound);
  } else {
    return Math.random();
  }
};

OOP_METHOD(Global, gettimeofday) = function ()
{
  return (new Date()).getTime();
};

OOP_METHOD(Global, time) = function ()
{
  var t1000 = this.gettimeofday();
  return Math.floor(t1000 / 1000);
};

OOP_METHOD(Global, escapeHtml) = function (str)
{
  var d = document.createElement('div');
  var text = document.createTextNode(str);
  d.appendChild(text);
  var ret = d.innerHTML;
  d = null;
  text = null;
  return ret;
};

OOP_METHOD(Global, zeroPrefix) = function (value, digits)
{
  if (+value === value) {
    var s = '';
    for (var i=0; i<digits; i++) s += '0';
    s += String(Number(value));
    return s.slice(s.length - digits);
  } else {
    return Number(value);
  }
};

OOP_METHOD(Global, encodeURIComponent) = function(str)
{
  return window.encodeURIComponent.apply(this, arguments);
};

OOP_METHOD(Global, decodeURIComponent) = function(str)
{
  var s = window.decodeURIComponent.apply(this, arguments);
  s = s.replace(/\+/g, ' ');
  return s;
};

OOP_METHOD(Global, arrayLength) = function (a)
{
  var count = 0;
  for (var i in a) {
    count++;
  }
  return count;
};

OOP_METHOD(Global, assignValues) = function (dst, src, fields)
{
  if (IS_OBJECT(fields)) {
    for (var i in fields) {
      var k = fields[i];
      if (IS_DEFINED(typeof src[k])) {
        dst[k] = src[k];
      }
    }
  } else {
    for (var j in src) {
      dst[j] = src[j];
    }
  }
};

OOP_METHOD(Global, quoteString) = function (str, q)
{
  if (str === null) return null;
  var s = String(str);
  s = s.replace(/\\/g,"\\\\");
  switch (q) {
  case "'":
    s = s.replace(/'/g,"\\'");
    break;
  case '"':
    s = s.replace(/"/g,'\\"');
    break;
  default:
    break;
  }
  s = s.replace(/\n/g,"\\n");
  s = s.replace(/\t/g,"\\t");
  s = s.replace(/\f/g,"\\f");
// FIXME: make the following line work
//    s = s.replace(/\b/g,"\\b");
  return q + s + q;
};

OOP_METHOD(Global, dumpVar) = function (obj, wantFunction, maxRecursion)
{
  var me = this;
  var maxLv = 64;
  var qq = '"';
  if (IS_DEFINED(maxRecursion)) maxLv = Number(maxRecursion);

  var f = function (o, lv) {
    if (lv > maxLv) {
      return {value:'(error)', error: 'Too many levels of recursion.'};
    }
    var t = typeof o;
    var v = new Object();
    var k;
    switch (t) {
    case 'number' :
    case 'boolean':
      v.value = String(o);
      break;
    case 'string' :
      v.value = me.quoteString(o, qq);
      break;
    case 'function':
      v.value = "/* function " + o + " */";
      v.isFunction = true;
      break;
    case 'undefined':
      v.value = "undefined";
      break;
    case 'object':
      // FIXME: Make it possible to dump arrays.
      if (o === null) {
        v.value = 'null';
      } else if (IS_ARRAY(o)) {
        // FIXME: handle the case where o is a pseudo hash.
        var lp1 = new PROJECT_PREFIX.text.ListPrinter('[', ',', ']');
        for (k in o) {
          var ret1 = f(o[k], lv + 1);
          if (!ret1.isFunction || wantFunction) {
            lp1.append(ret1.value);
          }
        }
        v.value = lp1.toString();
      } else {
        var lp2 = new PROJECT_PREFIX.text.ListPrinter('{', ',', '}');
        for (k in o) {
          var ret2 = f(o[k], lv + 1);
          if (!ret2.isFunction || wantFunction) {
            lp2.append(me.quoteString(k, qq) + ":" + ret2.value);
          }
        }
        v.value = lp2.toString();
      }
      break;
    default:
      FATAL('invalid data type: ' + t);
      v.value = null;
      break;
    }
    return v;
  };
  var r = f(obj, 0);
  return r.value;
};

OOP_METHOD(Global, invokeLater) = function (func, timeout)
{
  var t = (IS_DEFINED(timeout)) ? timeout : 0;
  if (IS_FUNCTION(func)) {
    setTimeout(func, t);
  }
};

#endif /* _JDISP_GLOBAL_JSH */
