#ifndef _JDISP_GUI_DRAGGABLE_JSH
#define _JDISP_GUI_DRAGGABLE_JSH

#define JDISP_GUI_DRAGGABLE_INIT     0
#define JDISP_GUI_DRAGGABLE_DRAGGING 10
#define JDISP_GUI_DRAGGABLE_ZINDEX   99999

OOP_CLASS(Draggable) = function(targetNode)
{
  this.targetNode = PROJECT_PREFIX.domUtils.getNode(targetNode);
  if (this.targetNode === null) {
    this.targetNode = document.documentElement;
  }
  this.nodes = [];
};

OOP_STATIC_FIELD(Draggable, DRAG_START) = 1;
OOP_STATIC_FIELD(Draggable, DRAG_MOVE)  = 2;
OOP_STATIC_FIELD(Draggable, DRAG_END)   = 3;
OOP_STATIC_METHOD(Draggable, defaultDragHandler) = function(state, rec, dx, dy)
{
  if (state === OOP_STATIC_FIELD(Draggable, DRAG_START)) {
    rec.srcPos = PROJECT_PREFIX.domUtils.getElementPosition(rec.dragTarget);
  } else if (state === OOP_STATIC_FIELD(Draggable, DRAG_MOVE)) {
    var x = rec.srcPos.x + dx;
    var y = rec.srcPos.y + dy;
    PROJECT_PREFIX.domUtils.setElementPosition(rec.dragTarget, x, y);
  } else if (state === OOP_STATIC_FIELD(Draggable, DRAG_END)) {
    rec.srcPos = null;
  } 
};

OOP_FIELD(Draggable, state) = JDISP_GUI_DRAGGABLE_INIT;
OOP_FIELD(Draggable, targetNode) = null;
OOP_FIELD(Draggable, nodes) = null;

OOP_METHOD(Draggable, add) = function (node, dragHandler)
{
  var n = PROJECT_PREFIX.domUtils.getNode(node);
  if (IS_NON_NULL_OBJECT(n)) {
    var me = this;
    var hh = IS_FUNCTION(dragHandler) ?
             dragHandler :
             OOP_STATIC_METHOD(Draggable, defaultDragHandler);
    var rec = {save:{}, node : n, dragHandler : hh, user: {}};
    rec.save.onmousedown = n.onmousedown;
    n.onmousedown = function (ev) {
      me._startDrag(ev, rec);
      PROJECT_PREFIX.domUtils.preventDefault(ev);
      PROJECT_PREFIX.domUtils.stopPropagation(ev);
      return false;
    };
    n.oncontextmenu = function (ev) {
      PROJECT_PREFIX.domUtils.preventDefault(ev);
      PROJECT_PREFIX.domUtils.stopPropagation(ev);
      return false;
    };
    PROJECT_PREFIX.domUtils.fixStaticPosition(n);
    this.nodes.push(rec);
  }
};

OOP_METHOD(Draggable, remove) = function (node)
{
  var n = PROJECT_PREFIX.domUtils.getNode(node);

  for (var i in this.nodes) {
    if (n === this.nodes[i]) {
      delete this.nodes[i];
    }
  }
};

OOP_METHOD(Draggable, _startDrag) = function (ev, rec)
{
  var me = this;
  var dat = PROJECT_PREFIX.domUtils.parseEventObject(ev);
  if (this.state !== JDISP_GUI_DRAGGABLE_INIT) {
    return;
  }
  rec.dat = dat;
  rec.dragTarget = dat.src;
  INVOKE_IF_SET(rec.dragHandler, OOP_STATIC_FIELD(Draggable, DRAG_START), rec);
  this.state = JDISP_GUI_DRAGGABLE_DRAGGING;
  var n = this.targetNode;
  // FIXME: handle the case where a third party library changes onmouseXXX.
  rec.save.onmousemove = n.onmousemove;
  rec.save.onmouseup = n.onmouseup;
  n.onmousemove = function (ev) {
    var dat2 = PROJECT_PREFIX.domUtils.parseEventObject(ev);
    var dx = dat2.position.x - dat.position.x;
    var dy = dat2.position.y - dat.position.y;
    rec.dat = dat2;
    INVOKE_IF_SET(rec.dragHandler, OOP_STATIC_FIELD(Draggable, DRAG_MOVE), rec, dx, dy);
    PROJECT_PREFIX.domUtils.preventDefault(ev);
    PROJECT_PREFIX.domUtils.stopPropagation(ev);
  };
  n.onmouseup = function (ev) {
    me._endDrag(ev, rec);
    PROJECT_PREFIX.domUtils.preventDefault(ev);
    PROJECT_PREFIX.domUtils.stopPropagation(ev);
 //   return false;
  };
};

OOP_METHOD(Draggable, _endDrag) = function (ev, rec)
{
  var dat = PROJECT_PREFIX.domUtils.parseEventObject(ev);
  if (this.state !== JDISP_GUI_DRAGGABLE_DRAGGING) {
    FATAL('internal error: invalid drag state');
  }
  rec.dat = dat;
  INVOKE_IF_SET(rec.dragHandler, OOP_STATIC_FIELD(Draggable, DRAG_END), rec);
  var n = this.targetNode;
  n.onmousemove = rec.save.onmousemove;
  n.onmouseup = rec.save.onmouseup;
  rec.save.onmousemove = null;
  rec.save.onmouseup = null;
  rec.srcPos = null;
  rec.dat = null;
  this.state = JDISP_GUI_DRAGGABLE_INIT;
};

#endif /* _JDISP_GUI_DRAGGABLE_JSH */
