#ifndef _JDISP_GUI_VIEWPORT_JSH
#define _JDISP_GUI_VIEWPORT_JSH

OOP_CLASS(Viewport) = function(node, x, y, blockWidth, blockHeight, draggable)
{
  var n = PROJECT_PREFIX.domUtils.getNode(node);
  this.targetNode = n || document.documentElement || document.body;
  IMPORT_ARGUMENT(x);
  IMPORT_ARGUMENT(y);
  IMPORT_ARGUMENT(blockWidth);
  IMPORT_ARGUMENT(blockHeight);
  IMPORT_ARGUMENT(draggable);
  this._initializeBlocks();
};

OOP_FIELD(Viewport, targetNode) = null;
OOP_FIELD(Viewport, x) = 0;
OOP_FIELD(Viewport, y) = 0;
OOP_FIELD(Viewport, viewMarginWidth) = 512;
OOP_FIELD(Viewport, viewMarginHeight) = 512;
OOP_FIELD(Viewport, blockWidth) = 128;
OOP_FIELD(Viewport, blockHeight) = 128;
OOP_FIELD(Viewport, blocks) = null;
OOP_FIELD(Viewport, spritePool) = null;
OOP_FIELD(Viewport, clock) = 0;
OOP_FIELD(Viewport, waitCount) = 0;
OOP_FIELD(Viewport, draggable) = null;

OOP_METHOD(Viewport, _bandEnds) = function (blockSize, offset, size, margin)
{
  var m = margin || 0;
  var startPos = Math.floor((offset - m) / blockSize);
  var endPos   = Math.floor((offset + m + size - 1) / blockSize);
  var sizeOut = endPos - startPos + 1;
  var ret = { startPos : startPos, endPos : endPos, size : sizeOut };
  return ret;
};

OOP_METHOD(Viewport, _initializeBlocks) = function ()
{
  var me = this;
  var n = this.targetNode;
  if (!IS_NON_NULL_OBJECT(n)) {
    // TODO: raise runtime exception
    throw new Exception("invalid target node");
  }
  n.style.overflow = 'hidden';
  n.onresize = function (ev) {
    me._initializeBlocks();
    me.update();
  };
  this.blocks = new Object();
  var es = PROJECT_PREFIX.domUtils.getElementSize(n);
  var bw = this.blockWidth;
  var bh = this.blockHeight;
  var vhn = 0;
  var vvn = 0;
  var i, ret;
  // TODO: optimize this.
  for (i=0; i<es.width; i++) {
    ret = this._bandEnds(this.blockWidth, i, es.width, this.viewMarginWidth);
    vhn = Math.max(vhn, ret.size);
  }
  for (i=0; i<es.height; i++) {
    ret = this._bandEnds(this.blockHeight, i, es.height, this.viewMarginHeight);
    vvn = Math.max(vvn, ret.size);
  }
  var vnn = 2 * vhn * vvn;
  LOG_DEBUG('Creating viewport of size ' + PROJECT_PREFIX.stringUtils.formatDimension(es));
  LOG_DEBUG('Creating sprite pool of size ' + vhn + 'x' + vvn);

  this.spritePool = new OOP_CLASS(SpritePool) (n, vnn);
  this.update();
};


OOP_METHOD(Viewport, moveTo) = function (x, y)
{
  if (x !== this.x || y !== this.y) {
    this.x = x;
    this.y = y;
    this.update();
  }
};

OOP_METHOD(Viewport, _updateBlock) = function (block, x, y)
{
  var sprite = block.sprite;
  PROJECT_PREFIX.domUtils.setElementPosition(sprite.node, x, y);
};

OOP_METHOD(Viewport, _prepareBlock) = function (block, x, y)
{
  var me = this;
  var sprite = block.sprite;
  sprite.node.innerHTML = '(' + block.key + ')';
  var s = sprite.node.style;
  s.position = 'absolute';
  var mw = 3, bw = 4, pw=5;
  mw = 0; bw = 0; pw=0;
  s.background = '#cccccc';
  s.border = 'solid ' + bw + 'px #808080';
  s.margin = mw + 'px ' + mw + 'px ' + mw + 'px ' + mw + 'px';
  s.padding = pw + 'px ' + pw + 'px ' + pw + 'px ' + pw + 'px';
  s.width = (this.blockWidth - 2 * bw - 2 * pw) + 'px';
  s.height = (this.blockHeight - 2 * bw - 2 * pw) + 'px';
  this._updateBlock(block, x, y);
  s.visibility = 'visible';
};

OOP_METHOD(Viewport, _disposeBlock) = function (block)
{
  var sprite = block.sprite;
  var s = sprite.node.style;
  s.visibility = 'hidden';
};

OOP_METHOD(Viewport, update) = function ()
{
  var startTime = PROJECT_PREFIX.global.gettimeofday();
  var clock = this.clock++;
  var es = PROJECT_PREFIX.domUtils.getElementSize(this.targetNode);
  var hor = this._bandEnds(this.blockWidth, this.x, es.width, this.viewMarginWidth);
  var ver = this._bandEnds(this.blockHeight, this.y, es.height, this.viewMarginHeight);
  var bw = this.blockWidth;
  var bh = this.blockHeight;
  var key, block, sprite;
  for (var i=ver.startPos; i<=ver.endPos; i++) {
    for (var j=hor.startPos; j<=hor.endPos; j++) {
      var x = bw * j - this.x;
      var y = bh * i - this.y;
      key = j + ',' + i;
      if (key in this.blocks) {
        block = this.blocks[key];
        this._updateBlock(block, x, y);
        block.clock = clock;
      } else {
        // expose the block
        sprite = this.spritePool.borrowItem();
        if (IS_NON_NULL_OBJECT(sprite)) {
          block = {sprite : sprite, clock: clock, initialClock: clock, key: key};
          block.remoteClock = 0;
          this._prepareBlock(block, x, y);
          this.blocks[key] = block;
//          log.append("appear: " + key + ' at ' + PROJECT_PREFIX.stringUtils.formatPoint({x:x, y:y}));
        } else {
          // TODO: handle error.
          FATAL("Failed to allocate sprite");
        }
      }
    }
  }
  for (key in this.blocks) {
    block = this.blocks[key];
    if (block.clock < clock) {
      // unexpose the block
      this._disposeBlock(block);
      this.spritePool.returnItem(block.sprite);
      block.sprite = null;
      delete this.blocks[key];
//      log.append("disappear: " + key);
    }
  }
#if 0
  var endTime = PROJECT_PREFIX.global.gettimeofday();
  PROJECT_PREFIX.logging.counter.add('viewport-update', endTime - startTime);
  log.append(PROJECT_PREFIX.logging.counter.toString());
#endif
};

OOP_METHOD(Viewport, _waitCountChanged) = function () {};

OOP_METHOD(Viewport, feed) = function (res)
{
  var i, block, rec, key, ix, iy;

  var me = this;
  var baseURL = res.baseURL || "/JSFB/image";
  baseURL += "/frame." + res.imageExt;

  var ret = new Object();
  ret.targets = new Array();
  ret.count = 0;

  for (i in res.blocks) {
    rec = res.blocks[i];
    ix = rec.ix;
    iy = rec.iy;
    key = ix + ',' + iy;
    if (key in this.blocks) {
      block = this.blocks[key];
      var clock = block.remoteClock;
      block.remoteClock = res.clock;
      if (rec.clock > clock) {
        ret.targets[i] = block;
        ret.count++;
      }
    }
  }

  this.waitCount += ret.count;
  this._waitCountChanged();

  for (i in ret.targets) {
    rec = res.blocks[i];
    ix = rec.ix;
    iy = rec.iy;
    key = ix + ',' + iy;
    if (key in this.blocks) {
      block = this.blocks[key];
      var url = baseURL + '?ix=' + ix + '&iy=' + iy + '&clock=' + rec.clock;
      var node = block.sprite.node;
      var img = node.firstChild;
      if (img === null || !IS_DEFINED(img.src)) {
        img = document.createElement('img');
        var f = function (e) {
          me.waitCount--;
          me._waitCountChanged();
        };
        img.alt = '';
        img.style.position = 'relative';
        img.style.top = '0';
        img.style.left = '0';
        img.onload = f;
        img.onerror = f;
        node.innerHTML = '';
        node.appendChild(img);
      }
      img.src = url;
    }
  }
};

#endif /* _JDISP_GUI_VIEWPORT_JSH */
