#ifndef _JDISP_LOGGING_COUNTER_JS
#define _JDISP_LOGGING_COUNTER_JS

OOP_CLASS(Counter) = function()
{
  this.data = new Object();
};

OOP_FIELD(Counter, data) = new Object();

OOP_METHOD(Counter, begin) = function (key)
{
  var rec = this.take(key);
  rec.begin++;
};

OOP_METHOD(Counter, end) = function (key)
{
  var rec = this.take(key);
  rec.end++;
  this.tick(key);
};

OOP_METHOD(Counter, tick) = function (key)
{
  var rec = this.take(key);
  rec.count++;
};

OOP_METHOD(Counter, add) = function (key, value)
{
  var rec = this.take(key);
  rec.sum += value;
  if (!rec.hasValue) {
    rec.max = value;
    rec.min = value;
  } else {
    rec.max = Math.max(rec.max, value);
    rec.min = Math.min(rec.min, value);
  }
  rec.hasValue = true;
  this.tick(key);
};



OOP_METHOD(Counter, take) = function (key)
{
  var rec = null;
  if (key in this.data) {
    rec = this.data[key];
  } else {
    rec = new Object();
    rec.begin = 0;
    rec.end = 0;
    rec.count = 0;
    rec.sum = 0;
    rec.hasValue = false;
    rec.max = 0;
    rec.min = 0;
    this.data[key] = rec;
  }

  return rec;
};

OOP_METHOD(Counter, toString) = function ()
{
  var lp = new PROJECT_PREFIX.text.ListPrinter('', ',', '');
  for (var key in this.data) {
    var rec = this.data[key];
    var statistics = '';
    if (rec.count > 0 && rec.hasValue) {
      var avg = rec.sum * 1.0 / rec.count;
      statistics = ',min=' + rec.min +
                   ',avg=' + avg.toFixed(3) +
                   ',max=' + rec.max ;
    }
    lp.append ( key + '=(' +
                rec.count + ',' +
                rec.begin + ',' + rec.end + statistics + ')');
  }
  return lp.toString();
};

#endif /* _JDISP_LOGGING_COUNTER_JS */
