#ifndef _JDISP_MACROS_JSH
#define _JDISP_MACROS_JSH

/*
 * primitive macros
 */
#define FATAL(msg) alert(msg)

#define STR_VEC2D(x, y) ("(" + Number(x) + "," + Number(y) + ")")
#define STR_POINT(pos) ((IS_OBJECT(pos)) ? STR_VEC2D(pos.x, pos.y) : '(not-a-point)')
#define STR_DIM(a) ((IS_OBJECT(a)) ? (a.width, a.height) : '(not-a-dimension)')

#define IS_FUNCTION(n) (typeof (n) === 'function')
#define IS_OBJECT(n) (typeof (n) === 'object')
#define IS_DEFINED(n) (typeof (n) !== 'undefined')
#define IS_STRING(n) (typeof (n) === 'string')
#define IS_NUMBER(n) (typeof (n) === 'number')
#define IS_NULL(n) ((n) === null)
#define IS_NON_NULL_OBJECT(n) (!IS_NULL(n) && IS_OBJECT(n))
#define IS_ARRAY(n) (n instanceof Array)

#define INVOKE_IF_SET_RAW(h, args...) if (IS_FUNCTION(h)) h(args)
#define INVOKE_IF_SET(h, args...) do { \
  INVOKE_IF_SET_RAW(h, args); \
} while (0)
#define DEFAULT_VALUE(h, d) ((IS_DEFINED(h) && !IS_NULL(h)) ? (h) : (d))
#define EVAL_RAW(txt) (function () {\
  if (window.execScript) {window.execScript(txt);} \
  else {window.eval(txt);} })()
#define EVAL(txt) do { EVAL_RAW(txt); } while(0)
#define SET_IF_DEF(name, val) if (IS_DEFINED(val)) name = val
#define IMPORT_ARGUMENT(name) SET_IF_DEF(this.name, name)

#define EMPTY_DICT() new Object()
#define EMPTY_ARRAY() new Array()

#define HAS_SUBSTRING(d, str) (String(d).indexOf(str) !== -1)
#ifdef DEBUG
# define LOG_DEBUG(msg) PROJECT_PREFIX.logging.stdout.append(msg)
#else
# define LOG_DEBUG(msg)
#endif
#define TRACE(msg) PROJECT_PREFIX.logging.stdout.append(__FILE__ + ': ' + msg)

#define COUNTER_TICK(sec) PROJECT_PREFIX.logging.counter.tick(sec)
#define COUNTER_BEGIN(sec) PROJECT_PREFIX.logging.counter.begin(sec)
#define COUNTER_END(sec) PROJECT_PREFIX.logging.counter.end(sec)

#endif /* _JDISP_MACROS_JSH */
